'''
Created on Dec 31, 2016

@author: nicolas
'''

import os

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import filter_testsuites_from_cli_args
from lemoncheesecake.project import find_project_file, Project
from lemoncheesecake.fixtures import FixtureRegistry, BuiltinFixture, load_fixtures_from_func
from lemoncheesecake.runner import run_testsuites
from lemoncheesecake.testsuite.filter import add_filter_args_to_cli_parser
from lemoncheesecake import reporting
from lemoncheesecake.exceptions import ProjectError, FixtureError, InvalidMetadataError,\
    ProgrammingError, LemonCheesecakeException, UserError, serialize_current_exception

def build_fixture_registry(project, cli_args):
    registry = FixtureRegistry()
    registry.add_fixture(BuiltinFixture("cli_args", lambda: cli_args))
    registry.add_fixture(BuiltinFixture("project_dir", lambda: project.get_project_dir()))
    for fixture in project.get_fixtures():
        registry.add_fixture(fixture)
    registry.check_dependencies()
    return registry

class RunCommand(Command):
    def get_name(self):
        return "run"

    def get_description(self):
        return "Run the tests"

    def add_cli_args(self, cli_parser):
        project_file = find_project_file()
        project = None
        default_reporting_backend_names = []
        if project_file:
            project = Project(project_file)
            default_reporting_backend_names = project.get_active_reporting_backend_names()

        add_filter_args_to_cli_parser(cli_parser)

        group = cli_parser.add_argument_group("Reporting")
        group.add_argument("--exit-error-on-failure", action="store_true",
            help="Exit with non-zero code if there is at least one non-passed test"
        )
        group.add_argument("--stop-on-failure", action="store_true",
            help="Stop tests execution on the first non-passed test"
        )
        group.add_argument("--report-dir", "-r", required=False, help="Directory where report data will be stored")
        group.add_argument("--reporting", nargs="+", default=default_reporting_backend_names,
            help="The list of reporting backends to use"
        )
        group.add_argument("--enable-reporting", nargs="+", default=[],
            help="The list of reporting backends to add (to base backends)"
        )
        group.add_argument("--disable-reporting", nargs="+", default=[],
            help="The list of reporting backends to remove (from base backends)"
        )

        if project:
            project.add_cli_extra_args(cli_parser)

    def run_cmd(self, cli_args):
        # Project initialization
        project_file = find_project_file()
        if not project_file:
            return "Cannot find project file"
        try:
            project = Project(project_file)
            testsuites = project.get_testsuites()
        except (ProjectError, ProgrammingError) as e:
            return str(e)
        except InvalidMetadataError as e:
            return "Invalid test/testsuite metadata has been found: %s" % e

        # Build fixture registry
        try:
            fixture_registry = build_fixture_registry(project, cli_args)
            fixture_registry.check_fixtures_in_testsuites(testsuites)
        except FixtureError as e:
            return "Cannot run tests: %s" % e

        reporting_backends = {
            backend.name: backend for backend in
                project.get_reporting_backends(capabilities=reporting.CAPABILITY_REPORTING_SESSION, active_only=False)
        }
        default_report_dir_creation_callback = project.get_report_dir_creation_callback()
        before_run_hook = project.get_before_test_run_hook()
        after_run_hook = project.get_after_test_run_hook()

        testsuites = filter_testsuites_from_cli_args(testsuites, cli_args)

        # Set reporting backends
        selected_reporting_backends = set()
        for backend_name in cli_args.reporting + cli_args.enable_reporting:
            try:
                selected_reporting_backends.add(reporting_backends[backend_name])
            except KeyError:
                return "Unknown reporting backend '%s'" % backend_name
        for backend_name in cli_args.disable_reporting:
            try:
                selected_reporting_backends.discard(reporting_backends[backend_name])
            except KeyError:
                return "Unknown reporting backend '%s'" % backend_name

        # Create report dir
        if cli_args.report_dir:
            report_dir = cli_args.report_dir
            try:
                os.mkdir(report_dir)
            except Exception as e:
                return "Cannot create report directory: %s" % e
        else:
            project_dir = project.get_project_dir()
            try:
                report_dir = default_report_dir_creation_callback(project_dir)
            except UserError as e:
                return str(e)
            except Exception:
                return "Got an unexpected exception while creating report directory:%s" % \
                    serialize_current_exception(show_stacktrace=True)

        # Handle before run hook
        if before_run_hook:
            try:
                before_run_hook(report_dir)
            except UserError as e:
                return str(e)
            except Exception:
                return "Got an unexpected exception while running the before-run hook:%s" % \
                    serialize_current_exception(show_stacktrace=True)

        # Run tests
        try:
            is_successful = run_testsuites(
                testsuites, fixture_registry, selected_reporting_backends, report_dir,
                stop_on_failure=cli_args.stop_on_failure
            )
        except LemonCheesecakeException as e:
            return str(e)

        # Handle after run hook
        if after_run_hook:
            try:
                after_run_hook(report_dir)
            except UserError as e:
                return str(e)
            except Exception:
                return "Got an unexpected exception while running the after-run hook:%s" % (
                    serialize_current_exception(show_stacktrace=True)
                )

        if cli_args.exit_error_on_failure:
            return 0 if is_successful else 1
        else:
            return 0



'''
Created on Sep 8, 2016

@author: nicolas
'''

import inspect

from lemoncheesecake.exceptions import InvalidMetadataError, ProgrammingError, InternalError
from lemoncheesecake.utils import get_distincts_in_list, get_callable_args

TESTSUITE_HOOKS = "setup_test", "teardown_test", "setup_suite", "teardown_suite"

__all__ = (
    "Test", "TestSuite", "walk_testsuites", "walk_tests", "filter_testsuites"
)

class Test:
    def __init__(self, name, description, callback):
        self.parent_suite = None
        self.name = name
        self.description = description
        self.callback = callback
        self.tags = [ ]
        self.properties = {}
        self.links = [ ]

    def get_params(self):
        return get_callable_args(self.callback)

    def get_path(self):
        return self.parent_suite.get_path() + [self]

    def get_path_str(self, sep="."):
        return sep.join([s.name for s in self.get_path()])

    def get_inherited_paths(self):
        return list(map(lambda s: s.get_path_str(), self.get_path()))

    def get_inherited_descriptions(self):
        return list(map(lambda s: s.description, self.get_path()))

    def get_inherited_tags(self):
        tags = []
        for suite in self.get_path():
            tags.extend(suite.tags)
        return get_distincts_in_list(tags)

    def get_inherited_properties(self):
        properties = {}
        for suite in self.get_path():
            properties.update(suite.properties)
        return properties

    def get_inherited_links(self):
        links = []
        for suite in self.get_path():
            links.extend(suite.links)
        return get_distincts_in_list(links)

def _assert_valid_hook_name(hook_name):
    if hook_name not in TESTSUITE_HOOKS:
        raise InternalError("Invalid hook name '%s'" % hook_name)

class TestSuite:
    def __init__(self, obj, name, description):
        self.obj = obj
        self.parent_suite = None
        self.name = name
        self.description = description
        self.tags = []
        self.properties = {}
        self.links = []
        self.rank = 0
        self._hooks = {}
        self._tests = []
        self._sub_testsuites = []
        self._selected_test_names = []

    def add_hook(self, hook_name, func):
        _assert_valid_hook_name(hook_name)
        self._hooks[hook_name] = func

    def has_hook(self, hook_name):
        _assert_valid_hook_name(hook_name)
        return hook_name in self._hooks

    def get_hook(self, hook_name):
        _assert_valid_hook_name(hook_name)
        return self._hooks.get(hook_name)

    def get_hook_params(self, hook_name):
        hook = self.get_hook(hook_name)
        assert hook != None
        return get_callable_args(hook)

    def get_path(self):
        suites = [ self ]
        parent_suite = self.parent_suite
        while parent_suite != None:
            suites.insert(0, parent_suite)
            parent_suite = parent_suite.parent_suite
        return suites

    def get_path_str(self, sep="."):
        return sep.join([s.name for s in self.get_path()])

    def __str__(self):
        return self.get_path_str()

    def get_depth(self):
        depth = 0
        parent = self.parent_suite
        while parent:
            depth += 1
            parent = parent.parent_suite
        return depth

    def assert_test_is_unique_in_suite(self, test):
        try:
            next(t for t in self._tests if t.description == test.description)
        except StopIteration:
            pass
        else:
            raise InvalidMetadataError(
                "A test with description '%s' is already registered in test suite %s" % (test.description, self.get_path_str())
            )

        try:
            next(t for t in self._tests if t.name == test.name)
        except StopIteration:
            pass
        else:
            raise InvalidMetadataError(
                "A test with name '%s' is already registered in test suite %s" % (test.name, self.get_path_str())
            )

    def assert_sub_suite_is_unique_in_suite(self, sub_suite):
        try:
            next(s for s in self._sub_testsuites if s.description == sub_suite.description)
        except StopIteration:
            pass
        else:
            raise InvalidMetadataError(
                "A sub test suite with description '%s' is already registered in test suite %s" % (sub_suite.name, self.get_path_str())
            )

        try:
            next(s for s in self._sub_testsuites if s.name == sub_suite.name)
        except StopIteration:
            pass
        else:
            raise InvalidMetadataError(
                "A sub test suite with name '%s' is already registered in test suite %s" % (sub_suite.name, self.get_path_str())
            )

    def add_test(self, test):
        self.assert_test_is_unique_in_suite(test)
        self._tests.append(test)
        test.parent_suite = self
        self._selected_test_names.append(test.name)

    def add_sub_testsuite(self, sub_suite):
        self.assert_sub_suite_is_unique_in_suite(sub_suite)
        sub_suite.parent_suite = self
        self._sub_testsuites.append(sub_suite)

    def get_test(self, test_name):
        for test in self._tests:
            if test.name == test_name:
                return test
        raise ProgrammingError("unknown test '%s'" % test_name)

    def get_tests(self, filtered=True):
        if filtered:
            return list(filter(lambda t: self.is_test_selected(t), self._tests))
        else:
            return self._tests

    def get_sub_testsuites(self, filtered=True):
        if filtered:
            return list(filter(lambda s: s.has_selected_tests(deep=True), self._sub_testsuites))
        else:
            return self._sub_testsuites

    def get_suite_name(self):
        return self.name

    def get_suite_description(self):
        return self.description

    def get_fixtures(self, filtered=True, recursive=True):
        fixtures = []

        suite_setup = self.get_hook("setup_suite")
        if suite_setup:
            fixtures.extend(get_callable_args(suite_setup))

        for test in self.get_tests(filtered):
            fixtures.extend(test.get_params())
        if recursive:
            for sub_suite in self.get_sub_testsuites(filtered):
                fixtures.extend(sub_suite.get_fixtures())

        return get_distincts_in_list(fixtures)

    ###
    # Filtering methods
    ###

    def apply_filter(self, filter):
        self._selected_test_names = [ ]

        for test in self._tests:
            if filter.match_test(test, self):
                self._selected_test_names.append(test.name)

        for suite in self._sub_testsuites:
            suite.apply_filter(filter)

    def has_selected_tests(self, deep=True):
        if deep:
            if self._selected_test_names:
                return True

            for suite in self._sub_testsuites:
                if suite.has_selected_tests():
                    return True

            return False
        else:
            return bool(self._selected_test_names)

    def is_test_selected(self, test):
        return test.name in self._selected_test_names

def walk_testsuites(testsuites, testsuite_func=None, test_func=None):
    def do_walk(suite):
        if testsuite_func:
            testsuite_func(suite)
        if test_func:
            for test in suite.get_tests():
                test_func(test, suite)
        for sub_suite in suite.get_sub_testsuites():
            do_walk(sub_suite)
    for suite in testsuites:
        do_walk(suite)

def walk_tests(testsuites, func):
    walk_testsuites(testsuites, test_func=func)

def filter_testsuites(suites, filtr):
    filtered = []
    for suite in suites:
        suite.apply_filter(filtr)
        if suite.has_selected_tests():
            filtered.append(suite)
    return filtered



'''
Created on Sep 8, 2016

@author: nicolas
'''

import fnmatch
from functools import reduce

__all__ = ("Filter", "add_filter_args_to_cli_parser", "get_filter_from_cli_args")

NEGATIVE_FILTER_CHARS = "-^~"

def match_values(values, patterns):
    if not patterns:
        return True

    values = [value or "" for value in values] # convert None to ""

    for pattern in patterns:
        if pattern[0] in NEGATIVE_FILTER_CHARS:
            if not fnmatch.filter(values, pattern[1:]):
                return True
        else:
            if fnmatch.filter(values, pattern):
                return True
    return False

def match_keyvalues(keyvalues, patterns):
    if not patterns:
        return True

    for key, value in patterns:
        if key in keyvalues:
            if value[0] in NEGATIVE_FILTER_CHARS:
                if not fnmatch.fnmatch(keyvalues[key], value[1:]):
                    return True
            else:
                if fnmatch.fnmatch(keyvalues[key], value):
                    return True
    return False

def match_values_lists(lsts, patterns):
    return match_values(
        reduce(lambda x, y: list(x) + list(y), lsts, []), # make a flat list
        patterns
    )

class Filter:
    def __init__(self):
        self.paths = []
        self.descriptions = []
        self.tags = [ ]
        self.properties = []
        self.links = [ ]

    def is_empty(self):
        return not any([self.paths, self.descriptions, self.tags, self.properties, self.links])

    def match_test(self, test, suite):
        funcs = [
            lambda: match_values(test.get_inherited_paths(), self.paths),
            lambda: all(match_values(test.get_inherited_descriptions(), descs) for descs in self.descriptions),
            lambda: all(match_values(test.get_inherited_tags(), tags) for tags in self.tags),
            lambda: all(match_keyvalues(test.get_inherited_properties(), props) for props in self.properties),
            lambda: all(match_values_lists(test.get_inherited_links(), links) for links in self.links)
        ]
        return all(func() for func in funcs)

def add_filter_args_to_cli_parser(cli_parser):
    def property_value(value):
        splitted = value.split(":")
        if len(splitted) != 2:
            raise ValueError()
        return splitted

    group = cli_parser.add_argument_group("Filtering")
    group.add_argument("path", nargs="*", default=[], help="Filter on test/testsuite path (wildcard character '*' can be used)")
    group.add_argument("--desc", nargs="+", action="append", default=[], help="Filter on descriptions")
    group.add_argument("--tag", "-a", nargs="+", action="append", default=[], help="Filter on tags")
    group.add_argument("--property", "-m", nargs="+", type=property_value, action="append", default=[], help="Filter on properties")
    group.add_argument("--link", "-l", nargs="+", action="append", default=[], help="Filter on links (names and URLs)")

def get_filter_from_cli_args(cli_args):
    fltr = Filter()
    fltr.paths = cli_args.path
    fltr.descriptions = cli_args.desc
    fltr.tags = cli_args.tag
    fltr.properties = cli_args.property
    fltr.links = cli_args.link
    return fltr


'''
Created on Mar 27, 2017

@author: nicolas
'''

import json

from lemoncheesecake.utils import get_distincts_in_list
from lemoncheesecake.exceptions import method_not_implemented

class MatchResult:
    def __init__(self, outcome, description):
        self.outcome = outcome
        self.description = description

    def is_success(self):
        return self.outcome == True

    def is_failure(self):
        return self.outcome == False

def match_success(description=None):
    return MatchResult(True, description)

def match_failure(description):
    return MatchResult(False, description)

def match_result(outcome, description=None):
    return MatchResult(outcome, description)

class Matcher:
    def description(self, conjugate=False):
        method_not_implemented("description", self)

    def matches(self, actual):
        method_not_implemented("match", self)

class MatchExpected(Matcher):
    def __init__(self, expected):
        self.expected = expected

def serialize_value(value):
    return json.dumps(value, ensure_ascii=False)

def serialize_values(values):
    return ", ".join(map(serialize_value, values))

def got(value):
    return "Got %s" % value

def got_value(value):
    return got(serialize_value(value))

def got_values(values):
    return got(serialize_values(values))

def merge_match_result_descriptions(results):
    return ", ".join(
        get_distincts_in_list([result.description for result in results if result.description != None])
    )

def to_be(conjugate=False):
    return "is" if conjugate else "to be"

def to_have(conjugate=False):
    return "has" if conjugate else "to have"



'''
Created on Mar 29, 2016

@author: nicolas
'''

import os

from lemoncheesecake.exceptions import MethodNotImplemented, InvalidReportFile,\
    ProgrammingError, method_not_implemented
from lemoncheesecake.utils import object_has_method

__all__ = (
    "get_available_backends", "ReportingBackend", "ReportingSession",
    "save_report", "load_reports_from_dir", "load_report",
    "CAPABILITY_REPORTING_SESSION", "CAPABILITY_SAVE_REPORT", "CAPABILITY_LOAD_REPORT"
)

CAPABILITY_REPORTING_SESSION = 0x1
CAPABILITY_SAVE_REPORT = 0x2
CAPABILITY_LOAD_REPORT = 0x4

SAVE_AT_END_OF_TESTS = 1
SAVE_AT_EACH_TESTSUITE = 2
SAVE_AT_EACH_TEST = 3
SAVE_AT_EACH_FAILED_TEST = 4
SAVE_AT_EACH_EVENT = 5

class ReportingSession:
    def begin_tests(self):
        pass

    def end_tests(self):
        pass

    def begin_test_session_setup(self):
        pass

    def end_test_session_setup(self):
        pass

    def begin_test_session_teardown(self):
        pass

    def end_test_session_teardown(self):
        pass

    def begin_suite(self, testsuite):
        pass

    def begin_suite_setup(self, testsuite):
        pass

    def end_suite_setup(self, testsuite):
        pass

    def begin_suite_teardown(self, testsuite):
        pass

    def end_suite_teardown(self, testsuite):
        pass

    def end_suite(self, testsuite):
        pass

    def begin_test(self, test):
        pass

    def end_test(self, test, status):
        pass

    def bypass_test(self, test, status, status_details):
        pass

    def set_step(self, description):
        pass

    def log(self, level, content):
        pass

    def check(self, description, outcome, details=None):
        pass

class ReportingBackend:
    def is_available(self):
        return True

    def get_capabilities(self):
        capabilities = 0
        if object_has_method(self, "create_reporting_session"):
            capabilities |= CAPABILITY_REPORTING_SESSION
        if object_has_method(self, "save_report"):
            capabilities |= CAPABILITY_SAVE_REPORT
        if object_has_method(self, "load_report"):
            capabilities |= CAPABILITY_LOAD_REPORT
        return capabilities

#     def create_reporting_session(self, dir, report):
#         method_not_implemented("create_reporting_session", self)
#
#     def save_report(self, filename, report):
#         method_not_implemented("serialize_report", self)
#
#     def load_report(self, filename):
#         method_not_implemented("unserialize_report", self)

class FileReportSession(ReportingSession):
    def __init__(self, report_filename, report, save_func, save_mode):
        self.report_filename = report_filename
        self.report = report
        self.save_func = save_func
        self.save_mode = save_mode

    def save(self):
        self.save_func(self.report_filename, self.report)

    def _handle_code_end(self, is_failure):
        if (self.save_mode == SAVE_AT_EACH_TEST) or (self.save_mode == SAVE_AT_EACH_FAILED_TEST and is_failure):
            self.save()
            return

    def end_test_session_setup(self):
        self._handle_code_end(
            self.report.test_session_setup.has_failure() if self.report.test_session_setup else False
        )

    def end_test_session_teardown(self):
        self._handle_code_end(
            self.report.test_session_teardown.has_failure() if self.report.test_session_teardown else False
        )

    def end_suite_setup(self, testsuite):
        suite_data = self.report.get_suite(testsuite.name)
        self._handle_code_end(
            suite_data.suite_setup.has_failure() if suite_data.suite_setup else False
        )

    def end_suite_teardown(self, testsuite):
        suite_data = self.report.get_suite(testsuite.name)
        self._handle_code_end(
            suite_data.suite_teardown.has_failure() if suite_data.suite_teardown else False
        )

    def end_test(self, test, status):
        self._handle_code_end(test)

    def end_suite(self, testsuite):
        if self.save_mode == SAVE_AT_EACH_TESTSUITE:
            self.save()

    def log(self, level, content):
        if self.save_mode == SAVE_AT_EACH_EVENT:
            self.save()

    def check(self, description, outcome, details=None):
        if self.save_mode == SAVE_AT_EACH_EVENT:
            self.save()

    def end_tests(self):
        self.save()

class FileReportBackend(ReportingBackend):
    def __init__(self, save_mode=SAVE_AT_EACH_FAILED_TEST):
        self.save_mode = save_mode

    def get_report_filename(self):
        method_not_implemented("get_report_filename", self)

    def create_reporting_session(self, report_dir, report):
        return FileReportSession(
            os.path.join(report_dir, self.get_report_filename()), report, self.save_report, self.save_mode
        )

def get_available_backends():
    from lemoncheesecake.reporting.backends import ConsoleBackend, XmlBackend, JsonBackend, HtmlBackend

    return list(filter(lambda b: b.is_available(), [ConsoleBackend(), XmlBackend(), JsonBackend(), HtmlBackend()]))

def load_report(filename, backends=None):
    if backends == None:
        backends = get_available_backends()
    for backend in backends:
        if backend.get_capabilities() & CAPABILITY_LOAD_REPORT:
            try:
                return backend.load_report(filename), backend
            except InvalidReportFile:
                pass
    raise InvalidReportFile("Cannot find any suitable report backend to unserialize file '%s'" % filename)

def load_reports_from_dir(dirname, backends=None):
    reports = []
    for filename in [os.path.join(dirname, filename) for filename in os.listdir(dirname)]:
        if os.path.isfile(filename):
            try:
                report, backend = load_report(filename, backends)
                reports.append((report, backend))
            except InvalidReportFile:
                pass
    return reports

def save_report(filename, report, backend):
    if not backend.get_capabilities() & CAPABILITY_SAVE_REPORT:
        raise ProgrammingError("Reporting backend '%s' does not support save operation" % backend.name)
    backend.save_report(filename, report)


'''
Created on Mar 26, 2016

@author: nicolas
'''

import time
import re
from decimal import Decimal

from lemoncheesecake.consts import LOG_LEVEL_ERROR, LOG_LEVEL_WARN
from lemoncheesecake.utils import humanize_duration

__all__ = (
    "LogData", "CheckData", "AttachmentData", "UrlData", "StepData", "TestData",
    "TestSuiteData", "HookData", "Report"
)

TEST_STATUSES = "passed", "failed", "skipped"

# NB: it would be nicer to use:
# datetime.isoformat(sep=' ', timespec='milliseconds')
# unfortunately, the timespec argument is only available since Python 3.6

def format_timestamp(ts):
    ts = round(ts, 3)
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(ts)) + (".%03d" % (Decimal(repr(ts)) % 1 * 1000))

def parse_timestamp(s):
    m = re.compile("(.+)\.(\d+)").match(s)
    if not m:
        raise ValueError("s is not valid datetime representation with milliseconds precision")

    dt, milliseconds = m.group(1), int(m.group(2))

    return time.mktime(time.strptime(dt, "%Y-%m-%d %H:%M:%S")) + float(milliseconds) / 1000

class LogData:
    def __init__(self, level, message, ts):
        self.level = level
        self.message = message
        self.time = ts

    def has_failure(self):
        return self.level == LOG_LEVEL_ERROR

class CheckData:
    def __init__(self, description, outcome, details=None):
        self.description = description
        self.outcome = outcome
        self.details = details

    def has_failure(self):
        return self.outcome == False

class AttachmentData:
    def __init__(self, description, filename):
        self.description = description
        self.filename = filename

    def has_failure(self):
        return False

class UrlData:
    def __init__(self, description, url):
        self.description = description
        self.url = url

    def has_failure(self):
        return False

class StepData:
    def __init__(self, description):
        self.description = description
        self.entries = [ ]
        self.start_time = None
        self.end_time = None

    def has_failure(self):
        return len(list(filter(lambda entry: entry.has_failure(), self.entries))) > 0

class TestData:
    def __init__(self, name, description):
        self.name = name
        self.description = description
        self.tags = [ ]
        self.properties = {}
        self.links = [ ]
        self.status = None
        self.status_details = None
        self.steps = [ ]
        self.start_time = None
        self.end_time = None

    def has_failure(self):
        return len(list(filter(lambda step: step.has_failure(), self.steps))) > 0

class HookData:
    def __init__(self):
        self.steps = [ ]
        self.start_time = None
        self.end_time = None
        self.outcome = None

    def has_failure(self):
        return len(list(filter(lambda step: step.has_failure(), self.steps))) > 0

    def is_empty(self):
        return len(self.steps) == 0

class TestSuiteData:
    def __init__(self, name, description, parent=None):
        self.name = name
        self.description = description
        self.parent = parent
        self.tags = [ ]
        self.properties = {}
        self.links = [ ]
        self.suite_setup = None
        self.tests = [ ]
        self.sub_testsuites = [ ]
        self.suite_teardown = None

    def get_test(self, test_name):
        for test in self.tests:
            if test.name == test_name:
                return test

        for suite in self.sub_testsuites:
            test = suite.get_test(test_name)
            if test:
                return test

        return None

    def get_suite(self, suite_name):
        if self.name == suite_name:
            return self

        for sub_suite in self.sub_testsuites:
            suite = sub_suite.get_suite(suite_name)
            if suite:
                return suite

        return None

class ReportStats:
    def __init__(self, report):
        self.tests = 0
        self.test_statuses = {s: 0 for s in TEST_STATUSES}
        self.errors = 0
        self.checks = 0
        self.check_successes = 0
        self.check_failures = 0
        self.error_logs = 0
        self.warning_logs = 0

        if report.test_session_setup:
            if report.test_session_setup.has_failure():
                self.errors += 1
            self._walk_steps(report.test_session_setup.steps)

        if report.test_session_teardown:
            if report.test_session_teardown.has_failure():
                self.errors += 1
            self._walk_steps(report.test_session_teardown.steps)

        for suite in report.testsuites:
            self._walk_testsuite(suite)

    def _walk_steps(self, steps):
        for step in steps:
            for entry in step.entries:
                if isinstance(entry, CheckData):
                    self.checks += 1
                    if entry.outcome == True:
                        self.check_successes += 1
                    elif entry.outcome == False:
                        self.check_failures += 1
                if isinstance(entry, LogData):
                    if entry.level == LOG_LEVEL_WARN:
                        self.warning_logs += 1
                    elif entry.level == LOG_LEVEL_ERROR:
                        self.error_logs += 1

    def _walk_testsuite(self, suite):
        if suite.suite_setup:
            if suite.suite_setup.has_failure():
                self.errors += 1
            self._walk_steps(suite.suite_setup.steps)

        if suite.suite_teardown:
            if suite.suite_teardown.has_failure():
                self.errors += 1
            self._walk_steps(suite.suite_teardown.steps)

        for test in suite.tests:
            self.tests += 1
            if test.status != None:
                self.test_statuses[test.status] += 1
            self._walk_steps(test.steps)

        for sub_suite in suite.sub_testsuites:
            self._walk_testsuite(sub_suite)

class Report:
    def __init__(self):
        self.info = [ ]
        self.test_session_setup = None
        self.test_session_teardown = None
        self.testsuites = [ ]
        self.start_time = None
        self.end_time = None
        self.report_generation_time = None

    def add_info(self, name, value):
        self.info.append([name, value])

    def get_test(self, test_name):
        for suite in self.testsuites:
            test = suite.get_test(test_name)
            if test:
                return test

        return None

    def get_suite(self, suite_name):
        for suite in self.testsuites:
            if suite.name == suite_name:
                return suite
            sub_suite = suite.get_suite(suite_name)
            if sub_suite:
                return sub_suite

        return None

    def get_stats(self):
        return ReportStats(self)

    def serialize_stats(self):
        stats = self.get_stats()
        return (
            ("Start time", time.asctime(time.localtime(self.start_time))),
            ("End time", time.asctime(time.localtime(self.end_time)) if self.end_time else "n/a"),
            ("Duration", humanize_duration(self.end_time - self.start_time) if self.end_time else "n/a"),
            ("Tests", str(stats.tests)),
            ("Successful tests", str(stats.test_statuses["passed"])),
            ("Successful tests in %", "%d%%" % (float(stats.test_statuses["passed"]) / stats.tests * 100 if stats.tests else 0)),
            ("Failed tests", str(stats.test_statuses["failed"])),
            ("Skipped tests", str(stats.test_statuses["skipped"])),
            ("Errors", str(stats.errors))
        )



'''
Created on Jan 24, 2016

@author: nicolas
'''

import os
import sys
import traceback

from lemoncheesecake.runtime import initialize_runtime, get_runtime
from lemoncheesecake.utils import IS_PYTHON3, get_distincts_in_list
from lemoncheesecake.exceptions import AbortTest, AbortTestSuite, AbortAllTests, FixtureError, \
    UserError, serialize_current_exception

class _Runner:
    def __init__(self, testsuites, fixture_registry, reporting_backends, report_dir, stop_on_failure=False):
        self.testsuites = testsuites
        self.fixture_registry = fixture_registry
        self.reporting_backends = reporting_backends
        self.report_dir = report_dir
        self.stop_on_failure = stop_on_failure

    def get_fixtures_with_dependencies_for_scope(self, direct_fixtures, scope):
        fixtures = []
        for fixture in direct_fixtures:
            fixtures.extend(self.fixture_registry.get_fixture_dependencies(fixture))
        fixtures.extend(direct_fixtures)
        return [f for f in get_distincts_in_list(fixtures) if self.fixture_registry.get_fixture_scope(f) == scope]

    def get_fixtures_to_be_executed_for_session_prerun(self):
        fixtures = []
        for testsuite in self.testsuites:
            fixtures.extend(testsuite.get_fixtures())
        return self.get_fixtures_with_dependencies_for_scope(get_distincts_in_list(fixtures), "session_prerun")

    def get_fixtures_to_be_executed_for_session(self):
        fixtures = []
        for testsuite in self.testsuites:
            fixtures.extend(testsuite.get_fixtures())
        return self.get_fixtures_with_dependencies_for_scope(get_distincts_in_list(fixtures), "session")

    def get_fixtures_to_be_executed_for_testsuite(self, testsuite):
        return self.get_fixtures_with_dependencies_for_scope(testsuite.get_fixtures(recursive=False), "testsuite")

    def get_fixtures_to_be_executed_for_test(self, test):
        return self.get_fixtures_with_dependencies_for_scope(test.get_params(), "test")

    def run_setup_funcs(self, funcs, failure_checker):
        teardown_funcs = []
        for setup_func, teardown_func in funcs:
            if setup_func:
                try:
                    setup_func()
                except (Exception, KeyboardInterrupt) as e:
                    self.handle_exception(e)
                    break
                else:
                    if failure_checker():
                        break
                    else:
                        teardown_funcs.append(teardown_func)
            else:
                teardown_funcs.append(teardown_func)
        return teardown_funcs

    def run_teardown_funcs(self, teardown_funcs):
        count = 0
        for teardown_func in teardown_funcs:
            if teardown_func:
                try:
                    teardown_func()
                except (Exception, KeyboardInterrupt) as e:
                    self.handle_exception(e)
                else:
                    count += 1
            else:
                count += 1
        return count

    def get_fixture_as_funcs(self, fixture):
        return [
            lambda: self.fixture_registry.execute_fixture(fixture),
            lambda: self.fixture_registry.teardown_fixture(fixture)
        ]

    def get_setup_suite_as_func(self, suite):
        setup_suite = suite.get_hook("setup_suite")
        if setup_suite == None:
            return None

        param_names = suite.get_hook_params("setup_suite")
        def func():
            params = self.fixture_registry.get_fixture_results_as_params(param_names)
            setup_suite(**params)

        return func

    def handle_exception(self, excp, suite=None):
        if isinstance(excp, AbortTest):
            self.session.log_error(str(excp))
        elif isinstance(excp, AbortTestSuite):
            self.session.log_error(str(excp))
            self.abort_testsuite = suite
        elif isinstance(excp, AbortAllTests):
            self.session.log_error(str(excp))
            self.abort_all_tests = True
        elif isinstance(excp, KeyboardInterrupt):
            self.session.log_error("All tests have been interrupted manually by the user")
            self.abort_all_tests = True
        else:
            # FIXME: use exception instead of last implicit stacktrace
            stacktrace = traceback.format_exc()
            if not IS_PYTHON3:
                stacktrace = stacktrace.decode("utf-8", "replace")
            self.session.log_error("Caught unexpected exception while running test: " + stacktrace)

    def run_test(self, test, suite):
        ###
        # Checker whether the test must be executed or not
        ###
        if self.abort_testsuite:
            self.session.skip_test(test, "Cannot execute this test: the tests of this test suite have been aborted.")
            return

        if self.abort_all_tests:
            self.session.skip_test(test, "Cannot execute this test: all tests have been aborted.")
            return

        ###
        # Begin test
        ###

        self.session.begin_test(test)

        ###
        # Setup test (setup and fixtures)
        ###
        test_setup_error = False
        setup_teardown_funcs = []
        teardown_funcs = []

        setup_teardown_funcs.append([
            (lambda: suite.get_hook("setup_test")(test.name)) if suite.has_hook("setup_test") else None,
            (lambda: suite.get_hook("teardown_test")(test.name)) if suite.has_hook("teardown_test") else None
        ])
        setup_teardown_funcs.extend([
            self.get_fixture_as_funcs(f) for f in self.get_fixtures_to_be_executed_for_test(test)
        ])

        if len(list(filter(lambda p: p[0] != None, setup_teardown_funcs))) > 0:
            self.session.begin_test_setup()
            teardown_funcs = self.run_setup_funcs(
                setup_teardown_funcs, lambda: self.session.current_test_data.has_failure()
            )
            self.session.end_test_setup()
            if len(teardown_funcs) != len(setup_teardown_funcs):
                test_setup_error = True
        else:
            teardown_funcs = [p[1] for p in setup_teardown_funcs if p[1] != None]
        
        if self.stop_on_failure and test_setup_error:
            self.abort_all_tests = True

        ###
        # Run test:
        ###
        if not test_setup_error:
            test_func_params = self.fixture_registry.get_fixture_results_as_params(test.get_params())
            try:
                test.callback(**test_func_params)
            except (Exception, KeyboardInterrupt) as e:
                self.handle_exception(e, suite)

        ###
        # Teardown
        ###
        if len(list(filter(lambda f: f != None, teardown_funcs))) > 0:
            self.session.begin_test_teardown()
            self.run_teardown_funcs(teardown_funcs)
            self.session.end_test_teardown()

        if self.stop_on_failure and self.session.current_test_data.has_failure():
            self.abort_all_tests = True

        self.session.end_test()

    def run_testsuite(self, suite):
        ###
        # Begin suite
        ###
        self.session.begin_suite(suite)

        ###
        # Setup suite (testsuites and fixtures)
        ###
        teardown_funcs = []
        if not self.abort_all_tests:
            setup_teardown_funcs = []
            setup_teardown_funcs.extend([
                self.get_fixture_as_funcs(f) for f in self.get_fixtures_to_be_executed_for_testsuite(suite)
            ])
            setup_teardown_funcs.append([
                self.get_setup_suite_as_func(suite), suite.get_hook("teardown_suite")
            ])

            if len(list(filter(lambda p: p[0] != None, setup_teardown_funcs))) > 0:
                self.session.begin_suite_setup()
                teardown_funcs = self.run_setup_funcs(
                    setup_teardown_funcs, lambda: self.session.current_testsuite_data.suite_setup.has_failure()
                )
                self.session.end_suite_setup()
                if len(teardown_funcs) != len(setup_teardown_funcs):
                    self.abort_testsuite = suite
            else:
                teardown_funcs = [p[1] for p in setup_teardown_funcs if p[1] != None]
            
            if self.stop_on_failure and self.abort_testsuite:
                self.abort_all_tests = True

        ###
        # Run tests
        ###
        for test in suite.get_tests(filtered=True):
            self.run_test(test, suite)

        ###
        # Teardown suite
        ###
        if len(list(filter(lambda f: f != None, teardown_funcs))) > 0:
            self.session.begin_suite_teardown()
            self.run_teardown_funcs(teardown_funcs)
            if self.stop_on_failure and self.session.current_testsuite_data.suite_teardown.has_failure():
                self.abort_all_tests = True
            self.session.end_suite_teardown()
        
        # reset the abort suite flag
        if self.abort_testsuite:
            self.abort_testsuite = None

        ###
        # Run sub testsuites
        ###
        for sub_suite in suite.get_sub_testsuites(filtered=True):
            self.run_testsuite(sub_suite)

        ###
        # End of testsuite
        ###

        self.session.end_suite()

    def run_session(self):
        # initialize runtime & global test variables
        initialize_runtime(self.reporting_backends, self.report_dir)
        self.session = get_runtime()
        self.session.initialize_reporting_sessions()
        self.abort_all_tests = False
        self.abort_testsuite = None

        # init report information
        self.session.report.add_info("Command line", " ".join([os.path.basename(sys.argv[0])] + sys.argv[1:]))

        self.session.begin_tests()

        # setup test session
        setup_teardown_funcs = []
        teardown_funcs = []
        setup_teardown_funcs.extend([
            self.get_fixture_as_funcs(f) for f in self.get_fixtures_to_be_executed_for_session()
        ])

        if len(list(filter(lambda p: p[0] != None, setup_teardown_funcs))) > 0:
            self.session.begin_test_session_setup()
            teardown_funcs = self.run_setup_funcs(
                setup_teardown_funcs, lambda: self.session.report.test_session_setup.has_failure()
            )
            self.session.end_test_session_setup()
            if len(teardown_funcs) != len(setup_teardown_funcs):
                self.abort_all_tests = True
        else:
            teardown_funcs = [p[1] for p in setup_teardown_funcs if p[1] != None]

        # run testsuites
        for suite in self.testsuites:
            self.run_testsuite(suite)

        # teardown_test_session handling
        if len(list(filter(lambda f: f != None, teardown_funcs))) > 0:
            self.session.begin_test_session_teardown()
            self.run_teardown_funcs(teardown_funcs)
            self.session.end_test_session_teardown()

        self.session.end_tests()

    def run(self):
        executed_fixtures = []

        errors = []
        for fixture in self.get_fixtures_to_be_executed_for_session_prerun():
            try:
                self.fixture_registry.execute_fixture(fixture)
            except UserError:
                raise
            except (Exception, KeyboardInterrupt):
                errors.append("Got the following exception when executing fixture '%s' (scope 'session_prerun')%s" % (
                    fixture, serialize_current_exception(show_stacktrace=True)
                ))
                break
            executed_fixtures.append(fixture)

        if not errors:
            self.run_session()

        for fixture in executed_fixtures:
            try:
                self.fixture_registry.teardown_fixture(fixture)
            except UserError:
                raise
            except (Exception, KeyboardInterrupt):
                errors.append("Got the following exception on fixture '%s' teardown (scope 'session_prerun')%s" % (
                    fixture, serialize_current_exception(show_stacktrace=True)
                ))

        if errors:
            raise FixtureError("\n".join(errors))
        
        return self.session.is_successful()

def run_testsuites(testsuites, fixture_registry, reporting_backends, report_dir, stop_on_failure=False):
    """
    Run testsuites.

    - testsuites: a list of already loaded testsuites (see lemoncheesecake.loader.load_testsuites)
    - reporting_backends: instance of reporting backends that will be used to report test results
    - report_dir: an existing directory where report files will be stored
    """
    runner = _Runner(testsuites, fixture_registry, reporting_backends, report_dir, stop_on_failure)
    return runner.run() # TODO: return a Report instance instead



'''
Created on Mar 28, 2017

@author: nicolas
'''

from lemoncheesecake.matching.base import Matcher, match_success, match_failure, match_result, \
    got_value, to_be, merge_match_result_descriptions

__all__ = (
    "all_of", "any_of", "anything", "something", "existing", "is_", "not_"
)

class AllOf(Matcher):
    def __init__(self, matchers):
        self.matchers = matchers

    def description(self, conjugate=False):
        return " and ".join([matcher.description(conjugate) for matcher in self.matchers])

    def matches(self, actual):
        results = []
        for matcher in self.matchers:
            result = matcher.matches(actual)
            if result.is_failure():
                return result
            results.append(result)

        return match_success(merge_match_result_descriptions(results))

def all_of(*matchers):
    """Test if all matchers match (logical AND between matchers)."""
    return AllOf(map(is_, matchers))

class AnyOf(Matcher):
    def __init__(self, matchers):
        self.matchers = matchers

    def description(self, conjugate=False):
        return " or ".join([matcher.description(conjugate) for matcher in self.matchers])

    def matches(self, actual):
        results = []
        for matcher in self.matchers:
            match = matcher.matches(actual)
            if match.is_success():
                return match
            results.append(match)

        return match_failure(merge_match_result_descriptions(results))

def any_of(*matchers):
    """Test if at least one of the matcher match (logical OR between matchers)"""
    return AnyOf(map(is_, matchers))

class Anything(Matcher):
    def __init__(self, wording="anything"):
        self.wording = wording

    def description(self, conjugate=False):
        return "%s %s" % (to_be(conjugate), self.wording)

    def matches(self, actual):
        return match_success(got_value(actual))

def anything():
    """Matches anything (always success, whatever the actual value)"""
    return Anything()

def something():
    """Same thing as the 'anything' matcher but use 'something' in the matcher description"""
    return Anything(wording="something")

def existing():
    """Same thing as the 'anything' matcher but use 'existing' in the matcher description"""
    return Anything(wording="existing")

def is_(matcher):
    """If the function argument is not an instance of Matcher, wrap it into
    a matcher using equal_to, otherwise return the matcher argument as-is"""
    from lemoncheesecake.matching.matchers.value import equal_to
    return matcher if isinstance(matcher, Matcher) else equal_to(matcher)

class Not(Matcher):
    def __init__(self, matcher):
        self.matcher = matcher

    def description(self, conjugate=False):
        return "not %s" % self.matcher.description(conjugate)

    def matches(self, actual):
        result = self.matcher.matches(actual)
        return match_result(not result.outcome, result.description)

def not_(matcher):
    """Negates the matcher in argument"""
    return Not(is_(matcher))


'''
Created on Mar 19, 2016

@author: nicolas
'''

from __future__ import print_function

import sys

from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSession
from lemoncheesecake.utils import IS_PYTHON3, humanize_duration
from lemoncheesecake.reporting.backends import terminalsize

import colorama
from termcolor import colored

class LinePrinter:
    def __init__(self, terminal_width):
        self.terminal_width = terminal_width
        self.prev_len = 0

    def print_line(self, line, force_len=None):
        value_len = force_len if force_len else len(line)
        if not IS_PYTHON3:
            if type(line) is unicode:
                line = line.encode("utf-8")

        if value_len >= self.terminal_width - 1:
            line = line[:self.terminal_width - 5] + "..."
            value_len = len(line)

        sys.stdout.write("\r")
        sys.stdout.write(line)
        if self.prev_len > value_len:
            sys.stdout.write(" " * (self.prev_len - value_len))
        sys.stdout.flush()

        self.prev_len = value_len

    def new_line(self):
        self.prev_len = 0
        sys.stdout.write("\n")
        sys.stdout.flush()

    def erase_line(self):
        sys.stdout.write("\r")
        sys.stdout.write(" " * self.prev_len)
        sys.stdout.write("\r")
        self.prev_len = 0

class ConsoleReportingSession(ReportingSession):
    def __init__(self, terminal_width, show_test_full_path, report):
        self.terminal_width = terminal_width
        self.show_test_full_path = show_test_full_path
        self.report = report
        colorama.init()
        self.lp = LinePrinter(self.terminal_width)
        self.context = None
        self.custom_step_prefix = None
        self.current_suite = None

    def get_test_label(self, test):
        if self.show_test_full_path:
            return test.get_path_str()
        return test.name

    def begin_tests(self):
        self.previous_obj = None

    def begin_suite(self, testsuite):
        self.current_suite = testsuite
        self.current_test_idx = 1

        if not testsuite.has_selected_tests(deep=False):
            return

        if self.previous_obj:
            sys.stdout.write("\n")

        label = testsuite.get_path_str()
        label_len = len(label)
        max_width = min((self.terminal_width, 80))
        # -2 corresponds to the two space characters at the left and right of testsuite path + another character to avoid
        # an extra line after the testsuite line on Windows terminal having width <= 80
        padding_total = max_width - 3 - label_len if label_len <= (max_width - 3) else 0
        padding_left = padding_total // 2
        padding_right = padding_total // 2 + padding_total % 2
        sys.stdout.write("=" * padding_left + " " + colored(label, attrs=["bold"]) + " " + "=" * padding_right + "\n")
        self.previous_obj = testsuite

    def begin_suite_setup(self, testsuite):
        self.step_prefix = " => setup suite: "
        self.lp.print_line(self.step_prefix + "...")

    def begin_suite_teardown(self, testsuite):
        self.step_prefix = " => teardown suite: "
        self.lp.print_line(self.step_prefix + "...")

    def begin_test_session_setup(self):
        self.step_prefix = " => setup test session: "
        self.lp.print_line(self.step_prefix + "...")

    def begin_test_session_teardown(self):
        self.step_prefix = " => teardown test session: "
        self.lp.print_line(self.step_prefix + "...")

    def end_suite_setup(self, testsuite):
        self.lp.erase_line()
        self.custom_step_prefix = None

    end_suite_teardown = end_suite_setup

    def end_test_session_setup(self):
        self.lp.erase_line()
        self.custom_step_prefix = None

    end_test_session_teardown = end_test_session_setup

    def begin_test(self, test):
        self.step_prefix = " -- %2s # %s" % (self.current_test_idx, self.get_test_label(test))
        self.lp.print_line(self.step_prefix + "...")
        self.previous_obj = test

    def end_test(self, test, status):
        line = " %s %2s # %s" % (
            colored("OK", "green", attrs=["bold"]) if status == "passed" else colored("KO", "red", attrs=["bold"]),
            self.current_test_idx, self.get_test_label(test)
        )
        raw_line = "%s %2s # %s" % ("OK" if status == "passed" else "KO", self.current_test_idx, self.get_test_label(test))
        self.lp.print_line(line, force_len=len(raw_line))
        self.lp.new_line()
        self.current_test_idx += 1

    def bypass_test(self, test, status, status_details):
        line = " %s %2s # %s" % (
            colored("KO", "yellow", attrs=["bold"]),
            self.current_test_idx, self.get_test_label(test)
        )
        raw_line = "%s %2s # %s" % ("KO", self.current_test_idx, self.get_test_label(test))
        self.lp.print_line(line, force_len=len(raw_line))
        self.lp.new_line()
        self.current_test_idx += 1

    def set_step(self, description):
        self.lp.print_line("%s (%s...)" % (self.step_prefix, description))

    def log(self, content, level):
        pass

    def end_tests(self):
        report = self.report
        stats = report.get_stats()

        print()
        print(colored("Statistics", attrs=["bold"]), ":")
        print(" * Duration: %s" % humanize_duration(report.end_time - report.start_time))
        print(" * Tests: %d" % stats.tests)
        print(" * Successes: %d (%d%%)" % (
            stats.test_statuses["passed"], float(stats.test_statuses["passed"]) / stats.tests * 100 if stats.tests else 0)
        )
        print(" * Failures: %d" % (stats.test_statuses["failed"]))
        if stats.test_statuses["skipped"]:
            print(" * Skipped: %d" % (stats.test_statuses["skipped"]))
        print()

class ConsoleBackend(ReportingBackend):
    name = "console"

    def __init__(self):
        width, height = terminalsize.get_terminal_size()
        self.terminal_width = width
        self.show_test_full_path = False

    def create_reporting_session(self, report_dir, report):
        return ConsoleReportingSession(self.terminal_width, self.show_test_full_path, report)



'''
Created on Feb 14, 2017

@author: nicolas
'''

from __future__ import print_function
from functools import reduce

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.display import print_table
from lemoncheesecake.cli.utils import filter_testsuites_from_cli_args
from lemoncheesecake.testsuite import add_filter_args_to_cli_parser, walk_testsuites
from lemoncheesecake.project import find_project_file, Project
from lemoncheesecake.exceptions import ProjectError, ProgrammingError

class StatsCommand(Command):
    def get_name(self):
        return "stats"

    def get_description(self):
        return "Display statistics about the project's tests"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Display")
        self.add_color_cli_args(group)
        add_filter_args_to_cli_parser(cli_parser)

    def run_cmd(self, cli_args):
        self.process_color_cli_args(cli_args)

        project_file = find_project_file()
        if not project_file:
            return "Cannot find project file"
        try:
            project = Project(project_file)
            suites = project.get_testsuites()
        except (ProjectError, ProgrammingError) as e:
            return str(e)

        suites = filter_testsuites_from_cli_args(suites, cli_args)

        class Stats:
            def __init__(self):
                self.tests_nb = 0
                self.testsuites_nb = 0
                self.non_empty_testsuites_nb = 0
                self.tags = {}
                self.properties = {}
                self.links = {}
        stats = Stats()

        def handle_test(test, suite):
            stats.tests_nb += 1
            for tag in test.get_inherited_tags():
                stats.tags[tag] = stats.tags.get(tag, 0) + 1
            for prop, value in test.get_inherited_properties().items():
                if prop not in stats.properties:
                    stats.properties[prop] = {}
                if value not in stats.properties[prop]:
                    stats.properties[prop][value] = 0
                stats.properties[prop][value] += 1
            for link in test.get_inherited_links():
                stats.links[link] = stats.links.get(link, 0) + 1

        def percent_of_tests(val):
            return "%2d%%" % (float(val) / stats.tests_nb * 100)

        def handle_suite(suite):
            stats.testsuites_nb += 1
            if suite.has_selected_tests(deep=False):
                stats.non_empty_testsuites_nb += 1

        walk_testsuites(suites, test_func=handle_test, testsuite_func=handle_suite)

        # Show tags
        lines = []
        for tag in sorted(stats.tags.keys(), key=lambda k: stats.tags[k], reverse=True):
            lines.append([self.bold(tag), stats.tags[tag], percent_of_tests(stats.tags[tag])])
        print_table(self.bold("Tags"), ["Tag", "Tests", "In %"], lines)

        # Show properties
        lines = []
        prop_names = sorted(
            stats.properties.keys(),
            key=lambda k: reduce(lambda x, y: x + y, stats.properties[k].values(), 0),
            reverse=True
        )
        for prop_name in prop_names:
            prop_values = sorted(
                stats.properties[prop_name].keys(),
                key=lambda k: stats.properties[prop_name][k],
                reverse=True
            )
            for prop_value in prop_values:
                lines.append([
                    self.bold(prop_name), self.bold(prop_value),
                    stats.properties[prop_name][prop_value],
                    percent_of_tests(stats.properties[prop_name][prop_value])
                ])
        print_table(self.bold("Properties"), ["Property", "Value", "Tests", "In %"], lines)

        # Show links
        lines = []
        for link in sorted(stats.links.keys(), key=lambda k: stats.links[k], reverse=True):
            lines.append([
                self.bold(link[1] or "-"), link[0], stats.links[link], percent_of_tests(stats.links[link])
            ])
        print_table(self.bold("Links"), ["Name", "URL", "Tests", "In %"], lines)

        summary = "Total: %s in %s" % (
            self.bold("%d tests" % stats.tests_nb),
            self.bold("%d testsuites" % stats.non_empty_testsuites_nb)
        )
        if stats.testsuites_nb > stats.non_empty_testsuites_nb:
            summary += " (+ %d empty suites)" % (stats.testsuites_nb - stats.non_empty_testsuites_nb)
        print(summary)
        print()

        return 0


from .run import RunCommand
from .bootstrap import BootstrapCommand
from .show import ShowCommand
from .fixtures import FixturesCommand
from .stats import StatsCommand

def get_commands():
    return [RunCommand(), BootstrapCommand(), ShowCommand(), FixturesCommand(), StatsCommand()]


'''
Created on Dec 10, 2016

@author: nicolas
'''

import os
import sys
import shutil
import imp
import inspect

from lemoncheesecake.testsuite import TestSuite
from lemoncheesecake.fixtures import Fixture
from lemoncheesecake.validators import MetadataPolicy
from lemoncheesecake.reporting import ReportingBackend, get_available_backends
from lemoncheesecake.reporting.reportdir import report_dir_with_archiving, archive_dirname_datetime
from lemoncheesecake.exceptions import ProjectError, UserError, serialize_current_exception
from lemoncheesecake.utils import get_resource_path, get_callable_args
from lemoncheesecake.importer import import_module

DEFAULT_REPORTING_BACKENDS = get_available_backends()

PROJECT_SETTINGS_FILE = "project.py"

def _find_file_going_up(filename, dirname):
    if os.path.exists(os.path.join(dirname, filename)):
        return os.path.join(dirname, filename)

    parent_dirname = os.path.dirname(dirname)
    if parent_dirname == dirname:
        return None # root directory has been reached

    return _find_file_going_up(filename, parent_dirname)

def find_project_file():
    filename = os.environ.get("LCC_PROJECT_FILE")
    if filename != None:
        return filename if os.path.exists(filename) else None

    filename = _find_file_going_up(PROJECT_SETTINGS_FILE, os.getcwd())
    return filename # filename can be None

def _check_class(klass=None):
    def wrapper(name, value):
        if klass:
            if not inspect.isclass(value) or not issubclass(value, klass):
                return "'%s' has an incorrect value, %s is not a subclass of %s" % (name, repr(value), klass)
        else:
            if not inspect.isclass(value):
                return "'%s' has an incorrect value, %s is not a class" % (name, repr(value))
        return None
    return wrapper

def _check_class_instance(klass):
    def wrapper(name, value):
        if not isinstance(value, klass):
            return "'%s' has an incorrect value, %s is not an instance of %s" % (name, value, klass)
        return None
    return wrapper

def _check_type(type_):
    def wrapper(name, value):
        if type(value) is not type_:
            return "'%s' has an incorrect value, '%s' is not a '%s'" % (name, value, type_)
        return None
    return wrapper

def _check_func(args_nb=None):
    def wrapper(name, value):
        if not callable(value):
            return "'%s' has an incorrect value, '%s' is not a function" % (name, value)
        if args_nb != None:
            args = get_callable_args(value)
            if len(args) != args_nb:
                return "'%s' function expect %s arguments, got %d" % (name, args_nb, len(args))
        return None
    return wrapper

def _check_report_backend(name, value):
    if not isinstance(value, ReportingBackend):
        return "%s does not inherit lemoncheesecake.reporting.ReportingBackend" % value
    return None

def _param_error(param_name, error_msg):
    raise ProjectError("Error with parameter '%s' of the project file: %s" % (param_name, error_msg))

class Project:
    def __init__(self, project_file):
        self._project_file = project_file
        self._project_dir = os.path.dirname(self._project_file)

        try:
            self._settings = import_module(self._project_file)
        except UserError as e:
            raise e # propagate UserError
        except Exception:
            raise ProjectError("Got an unexpected error while loading project:%s" % (
                serialize_current_exception()
            ))

        ###
        # Fetch parameters from project settings
        ###
        self._raw_params = {}
        def _(name, *args, **kwargs):
            self._raw_params[name] = self._get_param(name, *args, **kwargs)
        _("CLI_EXTRA_ARGS", _check_func(args_nb=1), required=False)
        _("REPORT_DIR_CREATION",
            _check_func(args_nb=1), required=False,
            default=lambda top_dir: report_dir_with_archiving(top_dir, archive_dirname_datetime)
        )
        _("TESTSUITES", _check_class_instance(TestSuite), is_list=True)
        _("FIXTURES", _check_class_instance(Fixture), is_list=True, required=False, default=[])
        _("REPORTING_BACKENDS",
            _check_class_instance(ReportingBackend), is_list=True, required=False, default=DEFAULT_REPORTING_BACKENDS
        )
        _("REPORTING_BACKENDS_ACTIVE",
            _check_type(str), is_list=True, required=False, default=None
        )
        _("METADATA_POLICY", _check_class_instance(MetadataPolicy), required=False, default=MetadataPolicy())
        _("RUN_HOOK_BEFORE_TESTS", _check_func(args_nb=1), required=False)
        _("RUN_HOOK_AFTER_TESTS", _check_func(args_nb=1), required=False)

        ###
        # Cross checks between parameters
        ###
        existing_backends = {backend.name: backend.is_available() for backend in self._raw_params["REPORTING_BACKENDS"]}
        if self._raw_params["REPORTING_BACKENDS_ACTIVE"] != None:
            for active_backend in self._raw_params["REPORTING_BACKENDS_ACTIVE"]:
                if active_backend in existing_backends:
                    if not existing_backends[active_backend]:
                        _param_error("REPORTING_BACKENDS_ACTIVE",
                            "backend '%s' is not available (available backends are: %s)" % (
                                active_backend, ", ".join([backend for backend, available in existing_backends.items() if available])
                        ))
                else:
                    _param_error("REPORTING_BACKENDS_ACTIVE",
                        "backend '%s' does not exist (backends are: %s)" % (
                            active_backend, ", ".join(existing_backends.keys())
                    ))
        else:
            self._raw_params["REPORTING_BACKENDS_ACTIVE"] = [
                backend for backend, is_available in existing_backends.items() if is_available
            ]

    def get_project_dir(self):
        return self._project_dir

    def _get_param(self, name, checker, is_list=False, is_dict=False, required=True, default=None):
        if not hasattr(self._settings, name):
            if required:
                _param_error(name, "required parameter is missing")
            else:
                return default

        value = getattr(self._settings, name)

        if is_list:
            if type(value) not in (list, tuple):
                _param_error(name, "parameter must be a list or a tuple")
            for v in value:
                error = checker(name, v)
                if error:
                    _param_error(name, error)
            value = list(value) # convert tuple in list
        elif is_dict:
            if type(value) != dict:
                _param_error(name, "parameter must be a dict")
            for v in value.values():
                error = checker(name, v)
                if error:
                    _param_error(name, error)

        else:
            error = checker(name, value)
            if error:
                _param_error(name, error)

        return value

    def get_cli_extra_args_callback(self):
        return self._raw_params["CLI_EXTRA_ARGS"]

    def add_cli_extra_args(self, cli_args_parser):
        callback = self.get_cli_extra_args_callback()
        if callback:
            group = cli_args_parser.add_argument_group("Project custom options")
            callback(group)

    def get_report_dir_creation_callback(self):
        return self._raw_params["REPORT_DIR_CREATION"]

    def get_testsuites(self, check_metadata_policy=True):
        suites = self._raw_params["TESTSUITES"]
        policy = self.get_metadata_policy()
        if check_metadata_policy and policy:
            policy.check_suites_compliance(suites)
        return suites

    def get_fixtures(self):
        return self._raw_params["FIXTURES"]

    def _get_reporting_backends(self):
        return self._raw_params["REPORTING_BACKENDS"]

    def get_reporting_backends(self, capabilities=0, active_only=False):
        if active_only:
            backends = filter(lambda b: b.name in self.get_active_reporting_backend_names(), self._get_reporting_backends())
        else:
            backends = self._get_reporting_backends()
        return list(filter(
            lambda b: b.get_capabilities() & capabilities == capabilities, backends
        ))

    def get_active_reporting_backend_names(self):
        return self._raw_params["REPORTING_BACKENDS_ACTIVE"]

    def is_reporting_backend_active(self, backend_name):
        return backend_name in self.get_active_reporting_backend_names()

    def get_metadata_policy(self):
        return self._raw_params["METADATA_POLICY"]

    def get_before_test_run_hook(self):
        return self._raw_params["RUN_HOOK_BEFORE_TESTS"]

    def get_after_test_run_hook(self):
        return self._raw_params["RUN_HOOK_AFTER_TESTS"]

def create_project(project_dir):
    p = os.path
    shutil.copyfile(get_resource_path(p.join("project", "template.py")), p.join(project_dir, PROJECT_SETTINGS_FILE))
    os.mkdir(p.join(project_dir, "tests"))
    os.mkdir(p.join(project_dir, "fixtures"))

def load_project(project_dir):
    return Project(os.path.join(project_dir, PROJECT_SETTINGS_FILE))



'''
Created on Jan 24, 2016

@author: nicolas
'''

import sys
import os.path
import time
import shutil

from lemoncheesecake.utils import humanize_duration
from lemoncheesecake.exceptions import LemonCheesecakeInternalError
from lemoncheesecake.consts import ATTACHEMENT_DIR, \
    LOG_LEVEL_DEBUG, LOG_LEVEL_ERROR, LOG_LEVEL_INFO, LOG_LEVEL_WARN
from lemoncheesecake.reporting import *

__all__ = "log_debug", "log_info", "log_warn", "log_warning", "log_error", "log_url", "set_step", \
    "prepare_attachment", "save_attachment_file", "save_attachment_content", \
    "add_report_info"

_runtime = None # singleton

def initialize_runtime(reporting_backends, report_dir):
    global _runtime
    _runtime = _Runtime(reporting_backends, report_dir)

def get_runtime():
    if not _runtime:
        raise LemonCheesecakeInternalError("Runtime is not initialized")
    return _runtime

class _Runtime:
    def __init__(self, reporting_backends, report_dir):
        self.reporting_backends = reporting_backends
        self.report_dir = report_dir
        self.attachments_dir = os.path.join(self.report_dir, ATTACHEMENT_DIR)
        self.attachment_count = 0
        self.report = Report()
        self.reporting_sessions = []
        self.step_lock = False
        self.default_step_description = None
        # pointers to report data parts
        self.current_testsuite_data = None
        self.current_test_data = None
        self.current_step_data_list = None
        self.current_step_data = None
        # pointers to running test/testsuite
        self.current_test = None
        self.current_testsuite = None
        # for test / testsuite hook / before/after all tests outcome
        self.has_pending_failure = False
        # global status
        self._is_success = True

    def initialize_reporting_sessions(self):
        for backend in self.reporting_backends:
            session = backend.create_reporting_session(self.report_dir, self.report)
            self.reporting_sessions.append(session)

    def for_each_reporting_sessions(self, callback):
        for session in self.reporting_sessions:
            callback(session)

    def _start_hook(self, ts):
        self.has_pending_failure = False
        hook_data = HookData()
        hook_data.start_time = ts
        return hook_data

    def _end_hook(self, hook_data, ts):
        if hook_data:
            hook_data.end_time = ts
            hook_data.outcome = not self.has_pending_failure

    def end_current_step(self, ts):
        if self.current_step_data:
            self.current_step_data.end_time = ts
            self.current_step_data = None

        # remove previous step from report data if it was empty
        if self.current_step_data_list and len(self.current_step_data_list[-1].entries) == 0:
            del self.current_step_data_list[-1]

    def begin_tests(self):
        self.report.start_time = time.time()
        self.for_each_reporting_sessions(lambda b: b.begin_tests())

    def end_tests(self):
        self.report.end_time = time.time()
        self.report.report_generation_time = self.report.end_time
        self.for_each_reporting_sessions(lambda b: b.end_tests())

    def begin_test_session_setup(self):
        self.report.test_session_setup = self._start_hook(time.time())
        self.current_step_data_list = self.report.test_session_setup.steps
        self.default_step_description = "Setup test session"

        self.for_each_reporting_sessions(lambda b: b.begin_test_session_setup())

    def end_test_session_setup(self):
        if self.report.test_session_setup.is_empty():
            self.report.test_session_setup = None
        else:
            now = time.time()
            self._end_hook(self.report.test_session_setup, now)
            self.end_current_step(now)
        self.for_each_reporting_sessions(lambda b: b.end_test_session_setup())

    def begin_test_session_teardown(self):
        self.report.test_session_teardown = self._start_hook(time.time())
        self.current_step_data_list = self.report.test_session_teardown.steps
        self.default_step_description = "Teardown test session"

        self.for_each_reporting_sessions(lambda b: b.begin_test_session_teardown())

    def end_test_session_teardown(self):
        if self.report.test_session_teardown.is_empty():
            self.report.test_session_teardown = None
        else:
            now = time.time()
            self._end_hook(self.report.test_session_teardown, now)
            self.end_current_step(now)
        self.for_each_reporting_sessions(lambda b: b.end_test_session_teardown())

    def begin_suite(self, testsuite):
        self.current_testsuite = testsuite
        suite_data = TestSuiteData(testsuite.name, testsuite.description, self.current_testsuite_data)
        suite_data.tags.extend(testsuite.tags)
        suite_data.properties.update(testsuite.properties)
        suite_data.links.extend(testsuite.links)
        if self.current_testsuite_data:
            self.current_testsuite_data.sub_testsuites.append(suite_data)
        else:
            self.report.testsuites.append(suite_data)
        self.current_testsuite_data = suite_data

        self.for_each_reporting_sessions(lambda b: b.begin_suite(testsuite))

    def begin_suite_setup(self):
        self.current_testsuite_data.suite_setup = self._start_hook(time.time())
        self.current_step_data_list = self.current_testsuite_data.suite_setup.steps
        self.default_step_description = "Setup suite"

        self.for_each_reporting_sessions(lambda b: b.begin_suite_setup(self.current_testsuite))

    def end_suite_setup(self):
        if self.current_testsuite_data.suite_setup.is_empty():
            self.current_testsuite_data.suite_setup = None
        else:
            now = time.time()
            self._end_hook(self.current_testsuite_data.suite_setup, now)
            self.end_current_step(now)
        self.for_each_reporting_sessions(lambda b: b.end_suite_setup(self.current_testsuite))

    def begin_suite_teardown(self):
        self.current_testsuite_data.suite_teardown = self._start_hook(time.time())
        self.current_step_data_list = self.current_testsuite_data.suite_teardown.steps
        self.default_step_description = "Teardown suite"

        self.for_each_reporting_sessions(lambda b: b.begin_suite_teardown(self.current_testsuite))

    def end_suite_teardown(self):
        if self.current_testsuite_data.suite_teardown.is_empty():
            self.current_testsuite_data.suite_teardown = None
        else:
            now = time.time()
            self.end_current_step(now)
            self._end_hook(self.current_testsuite_data.suite_teardown, now)
        self.for_each_reporting_sessions(lambda b: b.end_suite_teardown(self.current_testsuite))

    def end_suite(self):
        self.for_each_reporting_sessions(lambda b: b.end_suite(self.current_testsuite))
        self.current_testsuite_data = self.current_testsuite_data.parent
        self.current_testsuite = self.current_testsuite.parent_suite

    def begin_test(self, test):
        now = time.time()
        self.has_pending_failure = False
        self.current_test = test
        self.current_test_data = TestData(test.name, test.description)
        self.current_test_data.tags.extend(test.tags)
        self.current_test_data.properties.update(test.properties)
        self.current_test_data.links.extend(test.links)
        self.current_test_data.start_time = now
        self.current_testsuite_data.tests.append(self.current_test_data)
        self.for_each_reporting_sessions(lambda b: b.begin_test(test))
        self.current_step_data_list = self.current_test_data.steps
        self.default_step_description = test.description

    def begin_test_setup(self):
        self.set_step("Setup test")

    def end_test_setup(self):
        self.set_step(self.current_test.description)

    def begin_test_teardown(self):
        self.set_step("Teardown test")

    def end_test_teardown(self):
        pass

    def end_test(self):
        now = time.time()
        self.current_test_data.status = "failed" if self.has_pending_failure else "passed"
        self.current_test_data.end_time = now
        self.end_current_step(now)

        self.for_each_reporting_sessions(lambda b: b.end_test(self.current_test, self.current_test_data.status))

        self.current_test = None
        self.current_test_data = None
        self.current_step_data_list = None

    def _bypass_test(self, test, status, status_details):
        now = time.time()

        test_data = TestData(test.name, test.description)
        test_data.tags.extend(test.tags)
        test_data.properties.update(test.properties)
        test_data.links.extend(test.links)
        test_data.end_time = test_data.start_time = now
        test_data.status = status
        test_data.status_details = status_details
        self.current_testsuite_data.tests.append(test_data)

        self.for_each_reporting_sessions(lambda b: b.bypass_test(test, status, status_details))

    def skip_test(self, test, reason):
        self._is_success = False
        self._bypass_test(test, "skipped", reason)

    def create_step_if_needed(self, ts=None):
        if not self.current_step_data_list:
            self._set_step(self.default_step_description, ts or time.time())

    def _set_step(self, description, ts):
        self.end_current_step(ts)

        self.current_step_data = StepData(description)
        self.current_step_data.start_time = ts

        self.current_step_data_list.append(self.current_step_data)

        self.for_each_reporting_sessions(lambda b: b.set_step(description))

    def set_step(self, description, force_lock=False):
        if self.step_lock and not force_lock:
            return

        self._set_step(description, time.time())

    def log(self, level, content):
        now = time.time()
        self.create_step_if_needed(now)
        self.current_step_data.entries.append(LogData(level, content, now))
        self.for_each_reporting_sessions(lambda b: b.log(level, content))

    def log_debug(self, content):
        self.log(LOG_LEVEL_DEBUG, content)

    def log_info(self, content):
        self.log(LOG_LEVEL_INFO, content)

    def log_warn(self, content):
        self.log(LOG_LEVEL_WARN, content)

    def log_error(self, content):
        self._is_success = False
        self.has_pending_failure = True
        self.log(LOG_LEVEL_ERROR, content)

    def check(self, description, outcome, details=None):
        self.create_step_if_needed()
        self.current_step_data.entries.append(CheckData(description, outcome, details))

        if outcome == False:
            self._is_success = False
            self.has_pending_failure = True

        self.for_each_reporting_sessions(lambda b: b.check(description, outcome, details))

        return outcome

    def prepare_attachment(self, filename, description=None):
        self.create_step_if_needed()

        if not description:
            description = filename

        attachment_filename = "%04d_%s" % (self.attachment_count + 1, filename)
        self.attachment_count += 1
        if not os.path.exists(self.attachments_dir):
            os.mkdir(self.attachments_dir)
        self.current_step_data.entries.append(AttachmentData(description, "%s/%s" % (ATTACHEMENT_DIR, attachment_filename)))

        return os.path.join(self.attachments_dir, attachment_filename)
        # TODO: add hook for attachment

    def save_attachment_file(self, filename, description=None):
        target_filename = self.prepare_attachment(os.path.basename(filename), description)
        shutil.copy(filename, target_filename)

    def save_attachment_content(self, content, filename, description=None, binary_mode=False):
        target_filename = self.prepare_attachment(filename, description)

        fh = open(target_filename, "wb")
        fh.write(content if binary_mode else content.encode("utf-8"))
        fh.close()

    def log_url(self, url, description=None):
        self.create_step_if_needed()

        if not description:
            description = url

        self.current_step_data.entries.append(UrlData(description, url))
        # TODO: add hook for URL
    
    def is_successful(self):
        return self._is_success

def log_debug(content):
    """
    Log a debug level message.
    """
    get_runtime().log_debug(content)

def log_info(content):
    """
    Log a info level message.
    """
    get_runtime().log_info(content)

def log_warning(content):
    """
    Log a warning level message.
    """
    get_runtime().log_warn(content)

log_warn = log_warning

def log_error(content):
    """
    Log an error level message.
    """
    get_runtime().log_error(content)

def set_step(description):
    """
    Set a new step.
    """
    get_runtime().set_step(description)

def prepare_attachment(filename, description=None):
    """
    Prepare a attachment using a pseudo filename and an optional description.
    The function returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return get_runtime().prepare_attachment(filename, description)

def save_attachment_file(filename, description=None):
    """
    Save an attachment using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    get_runtime().save_attachment_file(filename, description)

def save_attachment_content(content, filename, description=None, binary_mode=False):
    """
    Save a given content as attachment using pseudo filename and optional description.
    """
    get_runtime().save_attachment_content(content, filename, description, binary_mode)

def log_url(url, description=None):
    """
    Log an URL.
    """
    get_runtime().log_url(url, description)

def add_report_info(name, value):
    report = get_runtime().report
    report.add_info(name, value)



import os.path

from lemoncheesecake.fixtures import load_fixtures_from_directory
from lemoncheesecake.testsuite import load_testsuites_from_directory
from lemoncheesecake.reporting.backends import ConsoleBackend, JsonBackend, XmlBackend, HtmlBackend
from lemoncheesecake.reporting.reportdir import report_dir_with_archiving, archive_dirname_datetime
from lemoncheesecake.validators import MetadataPolicy

###
# Variables
###
project_dir = os.path.dirname(__file__)

###
# Test suites
###
TESTSUITES = load_testsuites_from_directory(os.path.join(project_dir, "tests"))

###
# Fixtures
###
FIXTURES = load_fixtures_from_directory(os.path.join(project_dir, "fixtures"))

###
# Custom CLI arguments
###
def add_cli_args(cli_parser):
    pass
CLI_EXTRA_ARGS = add_cli_args

###
# Metadata policy
###
mp = MetadataPolicy()
METADATA_POLICY = mp

###
# Reporting
###
console_backend = ConsoleBackend()
json_backend = JsonBackend()
xml_backend = XmlBackend()
html_backend = HtmlBackend()

REPORTING_BACKENDS = console_backend, json_backend, xml_backend, html_backend
REPORTING_BACKENDS_ACTIVE = console_backend.name, json_backend.name, html_backend.name
REPORT_DIR_CREATION = lambda top_dir: report_dir_with_archiving(top_dir, archive_dirname_datetime)

###
# Test run hooks
###
def before_tests(report_dir):
    pass
RUN_HOOK_BEFORE_TESTS = before_tests

def after_tests(report_dir):
    pass
RUN_HOOK_AFTER_TESTS = after_tests



'''
Created on Feb 5, 2017

@author: nicolas
'''

import os.path as osp
import inspect

from lemoncheesecake.importer import get_matching_files, get_py_files_from_dir, strip_py_ext, import_module
from lemoncheesecake.exceptions import UserError, ProgrammingError, ImportTestSuiteError, InvalidMetadataError, serialize_current_exception
from lemoncheesecake.testsuite.core import Test, TestSuite, TESTSUITE_HOOKS

__all__ = "load_testsuite_from_file", "load_testsuites_from_files", "load_testsuites_from_directory", \
    "load_testsuite_from_class", "load_testsuites_from_classes"

def is_testsuite_class(obj):
    return inspect.isclass(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_testsuite

def is_test_method(obj):
    return inspect.ismethod(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_test

def is_test_function(obj):
    return inspect.isfunction(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_test

def load_test(obj):
    md = obj._lccmetadata
    test = Test(md.name, md.description, obj)
    test.tags.extend(md.tags)
    test.properties.update(md.properties)
    test.links.extend(md.links)
    return test

def load_test_from_method(method):
    return load_test(method)

def load_test_from_function(func):
    return load_test(func)

def _list_object_attributes(obj):
    return [getattr(obj, n) for n in dir(obj) if not n.startswith("__")]

def get_test_methods_from_class(obj):
    return sorted(filter(is_test_method, _list_object_attributes(obj)), key=lambda m: m._lccmetadata.rank)

def get_sub_suites_from_class(obj):
    return sorted(filter(is_testsuite_class, _list_object_attributes(obj)), key=lambda c: c._lccmetadata.rank)

def get_test_functions_from_module(mod):
    return filter(is_test_function, _list_object_attributes(mod))

def get_suite_classes_from_module(mod):
    return filter(is_testsuite_class, _list_object_attributes(mod))

def load_testsuite_from_class(klass):
    md = klass._lccmetadata
    try:
        inst = klass()
    except UserError as e:
        raise e # propagate UserError
    except Exception:
        raise ProgrammingError("Got an unexpected error while instanciating testsuite class '%s':%s" % (
            klass.__name__, serialize_current_exception()
        ))
    suite = TestSuite(inst, md.name, md.description)
    suite.tags.extend(md.tags)
    suite.properties.update(md.properties)
    suite.links.extend(md.links)
    suite.rank = md.rank

    for hook_name in TESTSUITE_HOOKS:
        if hasattr(inst, hook_name):
            suite.add_hook(hook_name, getattr(inst, hook_name))

    for test_method in get_test_methods_from_class(inst):
        suite.add_test(load_test_from_method(test_method))

    for sub_suite_klass in get_sub_suites_from_class(inst):
        suite.add_sub_testsuite(load_testsuite_from_class(sub_suite_klass))

    return suite

def load_testsuites_from_classes(klasses):
    return [load_testsuite_from_class(klass) for klass in klasses]

def load_testsuite_from_module(mod):
    # TODO: find a better way to workaround circular import
    from lemoncheesecake.testsuite.definition import get_metadata_next_rank

    suite_info = getattr(mod, "TESTSUITE")
    suite_name = inspect.getmodulename(inspect.getfile(mod))

    try:
        suite_description = suite_info["description"]
    except KeyError:
        raise InvalidMetadataError("Missing description in '%s' testsuite information" % mod.__file__)

    suite = TestSuite(None, suite_name, suite_description)
    suite.tags.extend(suite_info.get("tags", []))
    suite.properties.update(suite_info.get("properties", []))
    suite.links.extend(suite_info.get("links", []))
    suite.rank = suite_info.get("rank", get_metadata_next_rank())

    for hook_name in TESTSUITE_HOOKS:
        if hasattr(mod, hook_name):
            suite.add_hook(hook_name, getattr(mod, hook_name))

    for func in get_test_functions_from_module(mod):
        suite.add_test(load_test_from_function(func))

    sub_suites = []
    for klass in get_suite_classes_from_module(mod):
        sub_suites.append(load_testsuite_from_class(klass))
    sub_suites.sort(key=lambda suite: suite.rank)
    for sub_suite in sub_suites:
        suite.add_sub_testsuite(sub_suite)

    return suite

def load_testsuite_from_file(filename):
    """Get testsuite from Python module.

    A valid module is either:
    - a module containing a dict name 'TESTSUITE' with keys:
      - description (mandatory)
      - tags (optional)
      - properties (optional)
      - links (optional)
      - rank (optional)
    - a module that contains a testsuite class with the same name as the module name

    Raise a ImportTestSuiteError if the testsuite class cannot be imported.
    """
    try:
        mod = import_module(filename)
    except ImportError:
        raise ImportTestSuiteError(
            "Cannot import file '%s': %s" % (filename, serialize_current_exception(show_stacktrace=True))
        )

    if hasattr(mod, "TESTSUITE"):
        suite = load_testsuite_from_module(mod)
    else:
        mod_name = strip_py_ext(osp.basename(filename))
        try:
            klass = getattr(mod, mod_name)
        except AttributeError:
            raise ImportTestSuiteError("Cannot find class '%s' in '%s'" % (mod_name, mod.__file__))
        suite = load_testsuite_from_class(klass)
    return suite

def load_testsuites_from_files(patterns, excluding=[]):
    """
    Import testsuites from a list of files:
    - patterns: a mandatory list (a simple string can also be used instead of a single element list)
      of files to import; the wildcard '*' character can be used
    - exclude: an optional list (a simple string can also be used instead of a single element list)
      of elements to exclude from the expanded list of files to import
    Example: load_testsuites_from_files("test_*.py")
    """
    return [load_testsuite_from_file(f) for f in get_matching_files(patterns, excluding)]

def load_testsuites_from_directory(dir, recursive=True):
    """Find testsuite classes in modules found in dir.

    The function expect that:
    - each module (.py file) contains a class that inherits TestSuite
    - the class name must have the same name as the module name (if the module is foo.py
      the class must be named foo)
    If the recursive argument is set to True, sub testsuites will be searched in a directory named
    from the suite module: if the suite module is "foo.py" then the sub suites directory must be "foo".

    Raise ImportTestSuiteError if one or more testsuite cannot be imported.
    """
    if not osp.exists(dir):
        raise ImportTestSuiteError("Directory '%s' does not exist" % dir)
    suites = [ ]
    for filename in get_py_files_from_dir(dir):
        suite = load_testsuite_from_file(filename)
        if recursive:
            subsuites_dir = strip_py_ext(filename)
            if osp.isdir(subsuites_dir):
                for sub_suite in load_testsuites_from_directory(subsuites_dir, recursive=True):
                    suite.add_sub_testsuite(sub_suite)
        suites.append(suite)
    suites.sort(key=lambda suite: suite.rank)
    return suites



'''
Created on Mar 12, 2017

@author: nicolas
'''

from lemoncheesecake.testsuite import get_filter_from_cli_args, filter_testsuites
from lemoncheesecake.exceptions import UserError

def filter_testsuites_from_cli_args(suites, cli_args):
    if not any(suite.has_selected_tests() for suite in suites):
        raise UserError("No test is defined in your lemoncheesecake project.")

    filtr = get_filter_from_cli_args(cli_args)
    if not filtr.is_empty():
        suites = filter_testsuites(suites, filtr)
        if len(suites) == 0:
            raise UserError("The filter does not match any test")
    return suites
